#!/usr/bin/python

# Copyright: (c) 2018, Terry Jones <terry.jones@example.org>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = r'''
---
module: my_test

short_description: This is my test module

# If this is part of a collection, you need to use semantic versioning,
# i.e. the version is of the form "2.5.0" and not "2.4".
version_added: "1.0.0"

description: This is my longer description explaining my test module.

options:
    name:
        description: This is the message to send to the test module.
        required: true
        type: str
    new:
        description:
            - Control to demo if the result of this module is changed or not.
            - Parameter description can be a list as well.
        required: false
        type: bool
# Specify this value according to your collection
# in format of namespace.collection.doc_fragment_name
extends_documentation_fragment:
    - my_namespace.my_collection.my_doc_fragment_name

author:
    - Your Name (@yourGitHubHandle)
'''

EXAMPLES = r'''
# Pass in a message
- name: Test with a message
  my_namespace.my_collection.my_test:
    name: hello world

# pass in a message and have changed true
- name: Test with a message and changed output
  my_namespace.my_collection.my_test:
    name: hello world
    new: true

# fail the module
- name: Test failure of the module
  my_namespace.my_collection.my_test:
    name: fail me
'''

RETURN = r'''
# These are examples of possible return values, and in general should use other names for return values.
original_message:
    description: The original name param that was passed in.
    type: str
    returned: always
    sample: 'hello world'
message:
    description: The output message that the test module generates.
    type: str
    returned: always
    sample: 'goodbye'
'''


class Greenplum:
    def __init__(
            self,
            db=None,
            user=None,
            password=None,
            port=3000,
            master_host='localhost'

    ):

        self.db = db
        self.user = user
        self.password = password
        self.master_host = master_host
        self.port = port
        self.conn = psycopg2.connect(
            database=self.db,            host=self.master_host,
            user=self.user,
            password=self.password,
            port=self.port
        )
        self.conn.set_session(autocommit=True)
        self.cursor = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)

    def query(self, sql):
        try:
            self.cursor.execute(sql)
        except BaseException as error:
            return str(error)
        try:
            response_sqls = self.cursor.fetchall()
        except Exception:
            return False
        dict_result = []
        for row in response_sqls:
            dict_result.append(dict(row))
        return dict_result

    def resgroup_config(self):
        sql = 'SELECT * FROM gp_toolkit.gp_resgroup_config;'
        return self.query(sql)

    def resgroup_oid(self):
        sql = (
            'SELECT rolname, rsgname '
            'FROM pg_roles, pg_resgroup '
            'WHERE pg_roles.rolresgroup=pg_resgroup.oid;'
            )
        return self.query(sql)

    def create_rg(
            self,
            group_name=None,
            concurrency=None,
            cpu_rate_limit=None,
            memory_limit=None,
            memory_shared_quota=None,
            memory_spill_ratio=None
    ):
        params = {
            'CONCURRENCY': concurrency,
            'CPU_RATE_LIMIT': cpu_rate_limit,
            'MEMORY_LIMIT': memory_limit,
            'MEMORY_SHARED_QUOTA': memory_shared_quota,
            'MEMORY_SPILL_RATIO': memory_spill_ratio
        }
        sql = (
            f'CREATE RESOURCE GROUP '
            f'{group_name} WITH ('
            f'{", ".join("=".join((str(parameter_name),str(parameter_value))) for parameter_name,parameter_value in params.items() if parameter_value != "False")}'
            f');'
        )
        try:
            self.cursor.execute(sql)
        except BaseException as error:
            return {'status': False, 'msg': str(error)}
        return {'status': True, 'msg': self.cursor.rowcount}

    def present(
            self,
            group_name=None,
            concurrency=None,
            cpu_rate_limit=None,
            memory_limit=None,
            memory_shared_quota=None,
            memory_spill_ratio=None
    ):
        params = {
            'CONCURRENCY': concurrency,
            'CPU_RATE_LIMIT': cpu_rate_limit,
            'MEMORY_LIMIT': memory_limit,
            'MEMORY_SHARED_QUOTA': memory_shared_quota,
            'MEMORY_SPILL_RATIO': memory_spill_ratio
        }
        # Проверяем есть ли нужная группа
        find_group = False
        current_groups = self.resgroup_config()
        for current_group in current_groups:
            if current_group['groupname'] == group_name:
                find_group = True

        # Когда найдена группа, отправим запросы в базу
        changed = False
        if find_group:
            for parameter_name, parameter_value in params.items():
                current_group_parametrs = [
                    current_group for current_group in current_groups if current_group['groupname'] == group_name][0]
                current_group_parametr_value = str(current_group_parametrs[parameter_name.lower()])
                parameter_value = str(parameter_value)
                # Проверяем, что параметр передан в эту функцию
                if parameter_value != 'False':
                    # Выполнять, только если переданный параметр отличается от текущего
                    if parameter_value != current_group_parametr_value:
                        sql = (
                            f'ALTER RESOURCE GROUP {group_name} '
                            f'SET {parameter_name} {parameter_value};'
                        )
                        try:
                            self.cursor.execute(sql)
                        except BaseException as error:
                            return {'status': False, 'changed': changed, 'msg': str(error)}
                        changed = True
            return {'status': True, 'changed': changed}

        # Создаем нужную группу, если не наши на шаге выше.
        create = self.create_rg(
            group_name,
            concurrency,
            cpu_rate_limit,
            memory_limit,
            memory_shared_quota,
            memory_spill_ratio
        )
        if create['status']:
            if create['msg'] == -1:
                return {'status': True, 'changed': True}
            return {'status': True, 'changed': False}
        return {'status': False, 'msg': create['msg'], 'changed': False}

    def drop_rg(self, group_name):
        sql = f"DROP RESOURCE GROUP {group_name};"
        try:
            self.cursor.execute(sql)
        except BaseException as error:
            return {'status': False, 'msg': str(error), 'changed': False}
        return {'status': True, 'msg': self.cursor.rowcount, 'changed': True}

    def ater_role_to_group(self, rolname=None, rsgname=None):
        for oid in self.resgroup_oid():
            # Прверяем, что роль существует
            if oid['rolname'] == rolname and oid['rsgname'] == rsgname:
                return {'status': True, 'changed': False}
        sql = f'ALTER ROLE {rolname} RESOURCE GROUP {rsgname};'
        try:
            self.cursor.execute(sql)
        except BaseException as error:
            return {'status': False, 'msg': str(error), 'changed': False}
        if self.cursor.rowcount == -1:
            return {'status': True, 'changed': True}
        return {'status': True, 'changed': False}

    def close(self):
        self.cursor.close()
        self.conn.close()

from contextlib import suppress
from ansible.module_utils.basic import AnsibleModule

psycopg2 = None
with suppress(ImportError):
    import psycopg2
    import psycopg2.extras


def run_module():
    # define available arguments/parameters a user can pass to the module
    module_args = dict(
        state=dict(type='str', required=False, default=False),
        # new=dict(type='bool', required=False, default=False),
        db_port=dict(type='int', required=False, default=False),
        db_name=dict(type='str', required=True),
        db_user=dict(type='str', required=True),
        # db_password=dict(type='str', required=True, no_log=True),
        db_host=dict(type='str', required=False, default=False),
        resgroup_config=dict(type='str', required=False, default=False),
        resgroup_oid=dict(type='str', required=False, default=False),
        sql=dict(type='str', required=False, default=False),
        rsgname=dict(type='str', required=False, default=False),
        role=dict(type='str', required=False, default=False),
        concurrency=dict(type='str', required=False, default=False),
        cpu_rate_limit=dict(type='str', required=False, default=False),
        memory_limit=dict(type='str', required=False, default=False),
        memory_shared_quota=dict(type='str', required=False, default=False),
        memory_spill_ratio=dict(type='str', required=False, default=False)
    )

    result = dict(
        changed=False,
        resgroup_config=False
    )

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True
    )

    if psycopg2 is None:
        return module.fail_json("The psycopg2 module is required un host")

    if module.check_mode:
        module.exit_json(**result)

    gp = Greenplum(
        db=module.params['db_name'],
        user=module.params['db_user']
    )

    state = module.params['state']
    rsgname = module.params['rsgname']
    role = module.params['role']
    if state == 'present':
        if rsgname == 'False':
            module.fail_json('rsgname parameter is not set')
        if module.params['concurrency'] == 'False':
            module.fail_json('concurrency parameter is not set')
        if module.params['cpu_rate_limit'] == 'False':
            module.fail_json('cpu_rate_limit parameter is not set')
        if module.params['memory_limit'] == 'False':
            module.fail_json('memory_limit parameter is not set')

        create_status = gp.present(
            group_name=rsgname,
            concurrency=module.params['concurrency'],
            cpu_rate_limit=module.params['cpu_rate_limit'],
            memory_limit=module.params['memory_limit'],
            memory_shared_quota=module.params['memory_shared_quota'],
            memory_spill_ratio=module.params['memory_spill_ratio']
        )
        if create_status['status']:
            result = create_status
        else:
            return module.fail_json(f"create_status {create_status}")

    if state == 'absent':
        if rsgname == 'False':
            module.fail_json('rsgname parameter is not set')

        status = gp.drop_rg(group_name=rsgname)
        if status['status']:
            result = status
        else:
            return module.fail_json(f"absent {status['msg']}")

    if state == 'alter':
        if rsgname == 'False':
            module.fail_json('rsgname parameter is not set')
        if role == 'False':
            module.fail_json('rsgname parameter is not set')
        alter_status = gp.ater_role_to_group(rsgname=rsgname, rolname=role)
        if alter_status['status']:
            result = alter_status
        else:
            module.fail_json(alter_status['msg'])

    if 'sql' in module.params:
        if module.params['sql'] != "False":
            result['response_sql'] = gp.query(module.params['sql'])
    if 'resgroup_config' in module.params:
        if module.params['resgroup_config'] != "False":
            result['resgroup_config'] = gp.resgroup_config()
    if 'resgroup_oid' in module.params:
        if module.params['resgroup_oid'] != "False":
            result['resgroup_oid'] = gp.resgroup_oid()

    gp.close()
    module.exit_json(**result)


def main():
    run_module()


if __name__ == '__main__':
    main()
