/*
 * Инициализация хранилища данных на Greenplum
 * 1. Создаётся служебная схема dwh_V1_4.
 * 2. В схеме создаётся процедура генерации схем для проектов.
 *
 * Скрипт должен выполняться от имени superuser'а.
 *
 * Автор - Дмитрий Жучков
 * Дата создания файла - 01.05.2021
 *
 * Изменения в файле:
 * 28.05.2021 - ЖД - Замена BDP на DWH по всему тексту скрипта, переименование файла
 * 31.05.2021 - ЖД - Добавлено логирование всех действий по созданию проекта,
 *                   а также таблица для хранения списка созданных проектов.
 *
 */


-- Скрипт должен выполняться от имени пользователя dwh_admin, который является владельцем базы DWH
-- Для создания этого пользователя и базы нужно выполнить следующие две строки от имени superuser'а:
-- CREATE ROLE dwh_admin CREATEROLE LOGIN PASSWORD 'dwh_password';
-- CREATE DATABASE DWH WITH OWNER dwh_admin;



-- DROP SCHEMA dwh_v1_4;

CREATE SCHEMA if not exists dwh_v1_4 AUTHORIZATION dwh_admin;

set search_path = dwh_v1_4;

-- таблица шаблонных слоёв (схем) хранилища
-- drop table dwh_schema_template_t;

drop table if exists dwh_v1_4.dwh_schema_template_t;
create table if not exists dwh_v1_4.dwh_schema_template_t(schema_num int, schema_template varchar(60), schema_revision int, schema_description text, is_active boolean);

insert into dwh_v1_4.dwh_schema_template_t values(1, '{{ params.env_id }}_BUF', 1, 'Buffer - временное хранение входящих данных', true);
insert into dwh_v1_4.dwh_schema_template_t values(2, '{{ params.env_id }}_CON', 1, 'Conversion - таблицы переключевания', true);
insert into dwh_v1_4.dwh_schema_template_t values(3, '{{ params.env_id }}_DDS', 1, 'Detail Data Store - детальные данные хранилища', true);
insert into dwh_v1_4.dwh_schema_template_t values(4, '{{ params.env_id }}_ETL', 1, 'ETL - данные о ходе извлечения, загрузки и обработки данных', true);
insert into dwh_v1_4.dwh_schema_template_t values(5, '{{ params.env_id }}_EXP', 1, 'Export - данные для экспорта в сторонние системы', true);
insert into dwh_v1_4.dwh_schema_template_t values(6, '{{ params.env_id }}_PDM', 1, 'Private Data Marts - частные витрины данных для внутреннего использования', true);
insert into dwh_v1_4.dwh_schema_template_t values(7, '{{ params.env_id }}_REP', 1, 'Reports - данные для отчётов', true);
insert into dwh_v1_4.dwh_schema_template_t values(8, '{{ params.env_id }}_SDM', 1, 'Shared Data Marts - общедоступные витрины данных для повторного использования', true);

drop table if exists dwh_v1_4.dwh_project_t cascade;
create table dwh_v1_4.dwh_project_t(project_id varchar(50), schema_revision int, created_dttm timestamp default current_timestamp);
alter table dwh_v1_4.dwh_project_t add primary key (project_id);

drop table if exists dwh_v1_4.dwh_action_t;
create table dwh_v1_4.dwh_action_t(
	project_id varchar(50),
	action_author varchar(100) default current_user,
	action_stmt text,
	action_dttm timestamp default current_timestamp,
	action_status varchar(10),
	status_desc text
);
alter table dwh_v1_4.dwh_action_t add constraint fk_dwh_action_project foreign key (project_id) references dwh_v1_4.dwh_project_t(project_id);


CREATE OR REPLACE FUNCTION dwh_v1_4.f_is_empty(p_str character varying)
 RETURNS boolean LANGUAGE plpgsql SECURITY DEFINER
AS $function$
begin
	return (p_str is null) or (trim(p_str) = '');
end;
$function$
;


CREATE OR REPLACE FUNCTION dwh_v1_4.f_is_empty(p_int integer)
 RETURNS boolean LANGUAGE plpgsql SECURITY DEFINER
AS $function$
begin
	return (p_int is null);
end;
$function$
;


CREATE OR REPLACE FUNCTION dwh_v1_4.p_raise_exception(p_error_code character varying, p_error_message character varying, p_error_hint character varying DEFAULT ''::character varying, p_show_context boolean DEFAULT false)
 RETURNS integer LANGUAGE plpgsql SECURITY DEFINER
AS $$
declare

    v_context TEXT;
    v_global_show_context boolean := false;
    v_array varchar[];

begin

	/*
	 * Процедура p_raise_exception формирует исключение, присоединяя при необходимости к пользовательскому сообщению об ошибке
	 * дополнительную контекстную отладочную информацию.
	 *
	 * Идея процедуры в создании единой точки генерации исключений, с возможностью в дальнейшем развить это направление,
	 * например, логированием ошибок или особой реакцией на отдельные типы ошибок.
	 *
	 * Параметры:
	 * p_error_code - код ошибки, должен иметь вид MLxxx, где xхх - цифры,
	 * p_error_message - сообщение об ошибке,
	 * p_error_hint (необязательный) - подсказка по ошибке, по умолчанию - пустая строка,
	 * p_show_context (необязательный) - признак показа контекста, по умолчанию - false.
	 *
	 * Вместо явного указания третьего параметра процедуры можно использовать
	 * глобальную переменную пользовательской сессии DWH.SHOW_ERROR_CONTEXT/
	 * Любое непустое значение (not null и <> '') этой переменной интерпретируется как "Показывать контекст ошибок".
	 *
	 * Установка глобальной переменной выполняется запросом вида:
	 * set DWH.SHOW_ERROR_CONTEXT to 'true';
	 *
	 */

	-- Если есть глобальная переменная DWH.SHOW_ERROR_CONTEXT, то любое ее непустое значение интерпретируется как "Показывать контекст ошибок",
	v_global_show_context = not dwh_v1_4.f_is_empty(current_setting('DWH.SHOW_ERROR_CONTEXT', true));

	if p_show_context or v_global_show_context then
    	get diagnostics v_context = pg_context;

        -- исключаем из стека саму процедуру p_raise_exception и ее get diagnostics
        v_array = regexp_split_to_array(v_context, E'\n');
        v_context = v_array[3:100]; -- совсем не указать правую границу нельзя, поэтому указываем с запасом

   	end if;

	if length(p_error_code) != 0 and length(p_error_code) != 6 then
	    raise exception E'%: % %', 'В процедуру p_raise_exception передан неверный код ошибки', p_error_code, v_context
	        using errcode = 'DWH001', hint = 'Код ошибки должен иметь вид DWHxxx, где ххх - цифры.';
	end if;

    if f_is_empty(p_error_code) or f_is_empty(p_error_message) then
	    raise exception E'%: % % % %', 'В процедуре p_raise_exception параметры p_error_code и p_error_message не могут быть пустыми!', p_error_code, p_error_message, p_error_hint, p_show_context
	        using errcode = 'DWH002', hint = '';
    end if;

    -- Собственно генерация пользовательской ошибки
    if p_show_context or v_global_show_context then
        raise exception E'% \n-----context start-----\n %\n-----context end-----', p_error_message, v_context using errcode = p_error_code, hint = p_error_hint;
    else
        raise exception E'%', p_error_message using errcode = p_error_code, hint = p_error_hint;
    end if;

   return 0; -- псевдорезультат

end;
$$
;

CREATE OR REPLACE FUNCTION dwh_v1_4.f_replace_jinja_params(p_string character varying, p_params json DEFAULT '{}'::json)
 RETURNS character varying
 LANGUAGE plpgsql
 SECURITY DEFINER
AS $function$
declare

    v_string varchar := p_string;
    v_param_value character varying;
    r record;

begin

    for r in (select json_object_keys(p_params) as param) loop
        v_param_value := p_params ->> r.param;
        if not (f_is_empty(r.param)) and not (f_is_empty(v_param_value)) then
            v_string := replace(v_string, '{{ params.' || r.param || ' }}', v_param_value);
        end if;
    end loop;

    return v_string;

end;
$function$
;

create or replace function dwh_v1_4.p_check_db_user(p_db_user character varying, p_check_existence boolean default true)
 returns integer LANGUAGE plpgsql SECURITY DEFINER
AS $$
begin

   /*
    * Проверка идентификатора проектного пользователя.
    *
    * Второй параметр отвечает за проверку существования такой роли в базе данных.
    * Проверка существования актуальна всегда, поэтому по умолчанию она выполняется.
    */

	if dwh_v1_4.f_is_empty(p_db_user) then
        perform p_raise_exception('DWH012', format('Идентификатор пользователя должен быть указан!'));
    end if;

    if p_check_existence then
        perform from pg_user where lower(usename) = lower(p_db_user);
        if not found then
  	        perform p_raise_exception('DWH010', format('Неверный идентификатор, пользователь с именем "%s" в СУБД не существует или недоступен!', coalesce(p_db_user, '<null>')));
        end if;
    end if;

    return 0; -- псевдорезультат

end;
$$
;

CREATE OR REPLACE FUNCTION dwh_v1_4.p_execute_sql(p_project_id varchar(50), p_statement text)
	RETURNS text
	LANGUAGE plpgsql
AS $$
declare

	v_text text;

begin
	/*
	 * Фиксация в таблице dwh_action_t выполняемых запросов.
	 */

	if dwh_v1_4.f_is_empty(p_project_id) then
        perform p_raise_exception('DWH012', format('Параметр p_project_id должен быть указан!'));
    end if;

    if dwh_v1_4.f_is_empty(p_statement) then
        perform p_raise_exception('DWH012', format('Исполняемое SQL-выражение должно быть указано!'));
    end if;

    execute p_statement;

	insert into dwh_v1_4.dwh_action_t(project_id, action_stmt, action_status) values (p_project_id, p_statement,  'done');

	RETURN 'done';

EXCEPTION
    WHEN others then
        GET STACKED DIAGNOSTICS v_text := MESSAGE_TEXT;
        insert into dwh_v1_4.dwh_action_t(project_id, action_stmt, action_status, status_desc) values (p_project_id, p_statement,  'error', v_text);
      	RETURN 'ERROR: ' || v_text;
end;
$$
;

-- test
-- select dwh_v1_4.p_execute_sql('qweqwe', 'set search_path = dwh_v1_4');
-- select dwh_v1_4.p_execute_sql('qweqwe', 'qweqwe');
-- select * from dwh_action_t
-- select dwh_v1_4.p_execute_sql('qweqwe', 'set search_path = ''dwh_v1_4''');
-- select * from dwh_action_t


drop function if exists dwh_v1_4.p_init_new_project(varchar);

CREATE OR REPLACE FUNCTION dwh_v1_4.p_init_new_project(p_project_id varchar)
	RETURNS text
	LANGUAGE plpgsql
	VOLATILE
AS $$
declare
    v_sql text;
    v_schema_template varchar(60);
    v_schema_revision int;
    v_result text;
begin

   /*
    * Создание нового проекта по идентификатору проектного пользователя.
    */
    perform dwh_v1_4.p_execute_sql(p_project_id, 'set search_path = dwh_v1_4');

	if dwh_v1_4.f_is_empty(p_project_id) then
        perform p_raise_exception('DWH012', format('Идентификатор пользователя должен быть указан!'));
    end if;

    perform dwh_v1_4.p_execute_sql(p_project_id, format('create user %I with password %L', lower(p_project_id), p_project_id));
    perform dwh_v1_4.p_execute_sql(p_project_id, format('grant %I to dwh_admin' , lower(p_project_id)));
    perform dwh_v1_4.p_execute_sql(p_project_id, format('grant select on protocol pxf to %I;' , lower(p_project_id)));

    v_result = concat('User ', p_project_id,' created with same password!');

    select max(schema_revision) into v_schema_revision from dwh_v1_4.dwh_schema_template_t where is_active;

    perform dwh_v1_4.p_execute_sql(p_project_id, format('insert into dwh_v1_4.dwh_project_t(project_id, schema_revision) values (%L, %L)', p_project_id, v_schema_revision));

    for v_schema_template in
        select schema_template from dwh_v1_4.dwh_schema_template_t where is_active
    loop
        v_schema_template = lower(dwh_v1_4.f_replace_jinja_params(v_schema_template, format('{"env_id": "%I"}', p_project_id)::json));
        perform dwh_v1_4.p_execute_sql(p_project_id, format('create schema if not exists %I authorization %I', v_schema_template, p_project_id));
        v_result = v_result || E'\n' || concat('Schema ', v_schema_template, ' created!');
    end loop;

    v_result = v_result || E'\n' || dwh_v1_4.p_init_new_project_etl_schema(p_project_id);

    return v_result;

end;
$$
;

-- test
-- select p_init_new_project('first_project');
-- select * from dwh_v1_4.dwh_project_t;
-- select * from dwh_action_t order by action_dttm

drop function if exists dwh_v1_4.p_init_new_project_etl_schema(varchar);

create or replace function dwh_v1_4.p_init_new_project_etl_schema(p_project_id character varying)
  returns text language plpgsql security definer
as $$
declare
    v_sql text;
    v_result text;
begin

   /*
    * Инициализация ETL-схемы нового проекта по идентификатору проектного пользователя.
    */

	if f_is_empty(p_project_id) then
        perform p_raise_exception('DWH012', format('Идентификатор пользователя должен быть указан!'));
    end if;

    -- в скрипте параметризуется название схемы и имя проектного пользователя

    v_sql =
        'CREATE TABLE {{ params.etl_schema }}.flyway_schema_history_platforma ('
            'installed_rank integer NOT NULL,'
            'version character varying(50),'
            'description character varying(200) NOT NULL,'
            'type character varying(20) NOT NULL,'
            'script character varying(1000) NOT NULL,'
            'checksum integer,'
            'installed_by character varying(100) NOT NULL,'
            'installed_on timestamp without time zone DEFAULT now() NOT NULL,'
            'execution_time integer NOT NULL,'
            'success boolean NOT NULL'
        ') distributed by (installed_rank);'
        ''
		'ALTER TABLE {{ params.etl_schema }}.flyway_schema_history_platforma'
		'    ADD CONSTRAINT flyway_schema_history_platforma_pk PRIMARY KEY (installed_rank);'
		'CREATE INDEX flyway_schema_history_platforma_s_idx ON {{ params.etl_schema }}.flyway_schema_history_platforma USING btree (success);'

	    'ALTER TABLE {{ params.etl_schema }}.flyway_schema_history_platforma OWNER to dwh_admin;'
        'GRANT SELECT ON {{ params.etl_schema }}.flyway_schema_history_platforma to {{ params.project_user }};'
        ''
		'CREATE TABLE {{ params.etl_schema }}.flyway_schema_history_project ('
		'    installed_rank integer NOT NULL,'
		'    version character varying(50),'
		'    description character varying(200) NOT NULL,'
		'    type character varying(20) NOT NULL,'
		'    script character varying(1000) NOT NULL,'
		'    checksum integer,'
		'    installed_by character varying(100) NOT NULL,'
		'    installed_on timestamp without time zone DEFAULT now() NOT NULL,'
		'    execution_time integer NOT NULL,'
		'    success boolean NOT NULL'
		') distributed by (installed_rank);'
        ''
		'ALTER TABLE {{ params.etl_schema }}.flyway_schema_history_project'
		'    ADD CONSTRAINT flyway_schema_history_project_pk PRIMARY KEY (installed_rank);'
	 	'CREATE INDEX flyway_schema_history_project_s_idx ON {{ params.etl_schema }}.flyway_schema_history_project USING btree (success);'
		''
        'ALTER TABLE {{ params.etl_schema }}.flyway_schema_history_project OWNER to dwh_admin;'
        'GRANT SELECT ON {{ params.etl_schema }}.flyway_schema_history_project to {{ params.project_user }};'
        ''
		'CREATE VIEW {{ params.etl_schema }}.postgres_objects_v AS '
		'SELECT nsp.nspname AS schemaname,'
		'   cls.relname AS objectname,'
		'   rol.rolname AS objectowner,'
		'       CASE cls.relkind'
		'        WHEN ''r''::"char" THEN ''TABLE''::text'
		'        WHEN ''m''::"char" THEN ''MATERIALIZED_VIEW''::text'
		'        WHEN ''i''::"char" THEN ''INDEX''::text'
		'        WHEN ''s''::"char" THEN ''SEQUENCE''::text'
		'        WHEN ''v''::"char" THEN ''VIEW''::text'
		'        WHEN ''c''::"char" THEN ''TYPE''::text'
		'        ELSE (cls.relkind)::text'
		'        END AS objecttype'
		'   FROM ((pg_class cls'
		'     JOIN pg_roles rol ON ((rol.oid = cls.relowner)))'
		'     JOIN pg_namespace nsp ON ((nsp.oid = cls.relnamespace)))'
		'  WHERE ((nsp.nspname <> ALL (ARRAY[''information_schema''::name, ''pg_catalog''::name])) AND (nsp.nspname !~~ ''pg_toast%''::text) AND (rol.rolname = ''gpadmin''::name))'
		'  ORDER BY nsp.nspname, cls.relname;'
        ''
        'ALTER TABLE {{ params.etl_schema }}.postgres_objects_v OWNER to dwh_admin;'
        'GRANT SELECT ON {{ params.etl_schema }}.postgres_objects_v to {{ params.project_user }};'
        ''
		'CREATE TABLE if not exists {{ params.etl_schema }}.dyn_partition_log_t ('
		'	load_id int8 NULL,'
		'	job_id int8 NULL,'
		'	src_sys_id varchar(100) NULL,'
		'	dttm timestamp NULL,'
		'	dates varchar[] NULL,'
		'	status varchar(100) NULL,'
		'	part_value_list varchar[],'
		'	log_id serial not null'
		') distributed by (log_id);'
        ''
        'alter table {{ params.etl_schema }}.dyn_partition_log_t add constraint dyn_partition_log_t_pk primary key (log_id);'
    	''
        'ALTER TABLE {{ params.etl_schema }}.dyn_partition_log_t OWNER to dwh_admin;'
        'GRANT ALL ON {{ params.etl_schema }}.dyn_partition_log_t to {{ params.project_user }};'
        ''
        'GRANT ALL ON SEQUENCE {{ params.etl_schema }}.dyn_partition_log_t_log_id_seq TO {{ params.project_user }};'
        ''
		' create or replace function {{ params.etl_schema }}.f_get_sync_part(p_env character varying, p_dds_schema character varying, p_dds_table_name character varying) returns text'
		'     language plpgsql'
		' as $function$ '
		' declare '
		'     v_result text := '''';'
		'     v_sql text := '''';'
		' begin'
		'     WITH RECURSIVE r AS ('
		'     SELECT inhrelid::regclass as text_table_name,'
		'            parse_ident(inhrelid::regclass::varchar) table_name,'
		'            regexp_matches(pg_get_expr(chld.relpartbound, chld.oid, true), ''.*\(''''(.*)''''\)'') as part_value,'
		'            parse_ident(inhparent::regclass::varchar) as paren_table_name'
		'     FROM pg_inherits inh'
		'     JOIN pg_class chld on inh.inhrelid = chld.oid'
		'     WHERE  inhparent = (p_dds_schema||''.''||p_dds_table_name)::regclass'
		'     union'
		'     SELECT inhrelid::regclass as text_table_name,'
		'            parse_ident(inhrelid::regclass::varchar) table_name,'
		'            r.part_value||regexp_matches(pg_get_expr(chld.relpartbound, chld.oid, true), ''.*\(''''(.*)''''\)'') as part_value,'
		'            r.table_name as paren_table_name'
		'     FROM r, pg_inherits inh'
		'     JOIN pg_class chld on inh.inhrelid = chld.oid'
		'     WHERE inhparent = r.text_table_name'
		'     )'
		''
		'     select'
		'     format(''select mtl_prj_ctl.f_set_version_schema(''''%1$s'''');'', p_env)||chr(10)||'
		'     format(''delete from serving_part_list_t where serv_tab_id = f_get_serving_table_id(''''%1$s'''', ''''%2$s'''', ''''%3$s'''');'', p_env, p_dds_schema, p_dds_table_name)||chr(10)||'
		'     string_agg(distinct format(''insert into serving_part_list_t'
		'     (serv_tab_id, spec_id, table_schema, table_name, child_table_schema, child_table_name, part_value, time_start, time_end)'
		'     values'
		'     (f_get_serving_table_id(''''%6$s'''', ''''%7$s'''', ''''%8$s''''), f_get_serving_table_part_id(f_get_serving_table_id(''''%6$s'''', ''''%7$s'''', ''''%8$s''''),2), ''''%1$s'''', ''''%2$s'''', ''''%3$s'''', ''''%4$s'''', ''''%5$s'''', Null, Null);'','
		'     paren_table_name[1], paren_table_name[2], table_name[1], table_name[2], part_value, p_env, p_dds_schema, p_dds_table_name), chr(10)) into v_sql from r;'
		'     v_result = v_result||chr(10)||v_sql;'
		'     return v_result;'
		' end;'
		' $function$;'
        ''
		' alter function {{ params.etl_schema }}.f_get_sync_part(character varying, character varying, character varying) owner to dwh_admin;'
		' grant execute on function {{ params.etl_schema }}.f_get_sync_part(character varying, character varying, character varying) to {{ params.project_user }};'
		''
		' create or replace function {{ params.etl_schema }}.f_get_old_part_drop_script(p_dds_schema character varying, p_dds_table_name character varying, p_year character varying) returns text'
		'     language plpgsql'
		' as $function$'
		' declare'
		'     v_sql text := '''';'
		' begin'
		'    select string_agg(distinct format(''drop table %1$s;'', text_table_name), chr(10)) into v_sql'
		'     from ('
		'    select'
		'     inh_second.inhrelid::regclass as text_table_name,'
		'     (regexp_matches(pg_get_expr(chld.relpartbound, chld.oid, true), ''.*\(''''(.*)''''\)''))[1]::date as part_value'
		'     from pg_inherits inh_first'
		'     join pg_inherits inh_second on inh_first.inhrelid = inh_second.inhparent'
		'     join pg_class chld on inh_second.inhrelid = chld.oid'
		'     where inh_first.inhparent = (p_dds_schema||''.''||p_dds_table_name)::regclass) a'
		'     where to_char(a.part_value, ''YYYY'') = p_year;'
		'     return v_sql;'
		' end;'
		' $function$;'
        ''
		'alter function {{ params.etl_schema }}.f_get_old_part_drop_script(varchar, varchar, varchar) owner to dwh_admin;'
		'grant execute on function {{ params.etl_schema }}.f_get_old_part_drop_script(varchar, varchar, varchar) to {{ params.project_user }};'
        '';

    execute dwh_v1_4.f_replace_jinja_params(v_sql, format('{"project_user":"%1$s", "etl_schema":"%1$s_etl"}', p_project_id)::json);

    v_result = concat('ETL-schema for ', p_project_id,' initialized!');

    return v_result;

end;
$$
;

