#!/usr/bin/python

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    'status': ['preview'],
    'supported_by': 'Ivan Radaev'
}

DOCUMENTATION = '''
---
module: kafka_topic

short_description: This is my test module

version_added: "1.0.0"

description:
    - "Module to operate with kafka topics"

options:
    cluster:
        description:
            - Host and port for Kafka cluster broker connection (for e.g. localhost:6669)
        required: true
    topic:
        description:
            - Kafka topic name
        required: true
    state:
        description:
            - State of the given Kafka topic - present or absent
        required: false
        default: present
    partitions:
        description:
            - Number of partitions for the new Kafka topic
        required: false
        default: 1
    replication:
        description:
            - Number of replicas for the new Kafka topic
        required: false
        default: 1
    min_in_sync:
        description:
            - min.insync.replicas. When a producer sets acks to “all” (or “-1”), this configuration specifies the minimum number of replicas 
              that must acknowledge a write for the write to be considered successful. 
              If this minimum cannot be met, then the producer will raise an exception.
        required: false
        default: 1 
    retention_ms:
        description:
            - retention.ms. This configuration controls the maximum time we will retain a log before we will discard 
              old log segments to free up space if we are using the “delete” retention policy. 
              This represents an SLA on how soon consumers must read their data. If set to -1, no time limit is applied.
        required: false
        default: 604800000 
    retention_bytes:
        description:
            - retention.bytes. This configuration controls the maximum size a partition (which consists of log segments) 
              can grow to before we will discard old log segments to free up space if we are using the “delete” retention policy. 
              By default there is no size limit only a time limit. 
              Since this limit is enforced at the partition level, multiply it by the number of partitions to compute the topic retention in bytes.
        required: false
        default: -1        

author:
    - Radaev Ivan (@myservices.digital)
'''

EXAMPLES = '''
# Local console test
python -m kafka_topic '{"ANSIBLE_MODULE_ARGS":{"cluster":"rnd-infra-1-kafka-01.khd.lan:6667","topic":"test","state":"present"}}'

# Create topic
- name: Create
  kafka_topic:
    cluster: rnd-infra-1-kafka-01.khd.lan:6667
    topic: test
    partitions: 3
    replication: 3
    min_in_sync: 2

# Delete topic
- name: Delete
  kafka_topic:
    cluster: rnd-infra-1-kafka-01.khd.lan:6667
    topic: test
    state: absent
'''

RETURN = '''
message:
    description: The output message that describes result state for the given topic or error message
    type: str
    returned: always
'''

from ansible.module_utils.basic import AnsibleModule
from kafka.admin import KafkaAdminClient, NewTopic


def run_module():
    module_args = dict(
        cluster=dict(type='str', required=True),
        topic=dict(type='str', required=True),
        state=dict(type='str', required=False, default='present'),
        partitions=dict(type='int', required=False, default=1),
        replication=dict(type='int', required=False, default=1),
        min_in_sync=dict(type='str', required=False, default='1'),
        retention_ms=dict(type='str', required=False, default='604800000'),
        retention_bytes=dict(type='str', required=False, default='-1')
    )

    result = dict(
        changed=False,
        message='Error'
    )

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True
    )

    topic_state = module.params['state']
    topic_name = module.params['topic']
    topic_partitions = module.params['partitions']
    topic_replication_factor = module.params['replication']
    kafka_server = module.params['cluster']
    kafka_client_id = 'ansible'
    topic_advanced_cfg = {'min.insync.replicas': module.params['min_in_sync'], 'retention.ms': module.params['retention_ms'], 'retention.bytes': module.params['retention_bytes']}
    topic_list = []

    try:
        admin_client = KafkaAdminClient(bootstrap_servers=kafka_server, client_id=kafka_client_id)
        existing_topic_list = admin_client.list_topics()
    except Exception as error:
        module.fail_json(msg=str(error), **result)

    topic_exists = None
    if topic_name in existing_topic_list:
        topic_exists = True

    if topic_state == 'present':
        if not topic_exists:
            if not module.check_mode:
                topic_list.append(NewTopic(name=topic_name, num_partitions=topic_partitions, replication_factor=topic_replication_factor, topic_configs=topic_advanced_cfg))
                try:
                    admin_client.create_topics(new_topics=topic_list, validate_only=False)
                except Exception as error:
                    module.fail_json(msg=str(error), **result)
            result['message'] = 'Topic has been created'
            result['changed'] = True
        else:
            result['message'] = 'Topic already exists'
            result['changed'] = False
    elif topic_state == 'absent':
        if topic_exists:
            if not module.check_mode:
                topic_list.append(topic_name)
                try:
                    admin_client.delete_topics(topics=topic_list)
                except Exception as error:
                    module.fail_json(msg=str(error), **result)
            result['message'] = 'Topic has been marked for deletion'
            result['changed'] = True
        else:
            result['message'] = 'Topic is absent'
            result['changed'] = False
    else:
        module.fail_json(msg='Error, input variable "topic_state" have to be either set to "present" or "absent"', **result)

    module.exit_json(**result)


def main():
    run_module()


if __name__ == '__main__':
    main()
