from jinja2 import Environment, FileSystemLoader
import yaml
import configparser
import sys
import subprocess
import logging
import pathlib



class BuildFiles:

    def __init__(
        self,
        cluster_name = 'datapark',
        ansible_ssh_pass = 'password',
        templates_folder = 'templates',
        super_secret_pass = 'password',
        service_user_pass = 'password',
        path_out_file = 'result',
        start_hosts_file = 'start_hosts.yml',
        config_file = 'config.ini',
        ansible_hosts_template = 'hosts.yml.j2',
        ansible_vars_template = 'common_vars.yml.j2',
        clustertemplate_template = 'clustertemplate.json.j2',
        out_clustertemplate_name = 'clustertemplate.json',
        out_ansible_hosts_name = 'hosts',
        out_ansible_vars_name = 'common_vars.yml',
        ansible_sudo_pass = ''
    ) -> None:
        
        current_path = str(pathlib.Path(__file__).parent.absolute())
        self.templates_folder = f'{current_path}/{templates_folder}'
        self.path_out_file = f'{current_path}/{path_out_file}'
        self.start_hosts_file = f'{current_path}/{start_hosts_file}'
        self.config = f'{current_path}/{config_file}'
        self.cluster_name = cluster_name
        self.ansible_ssh_pass = ansible_ssh_pass
        self.super_secret_pass = super_secret_pass
        self.service_user_pass = service_user_pass
        self.ansible_hosts_template = ansible_hosts_template
        self.ansible_vars_template = ansible_vars_template
        self.clustertemplate_template = clustertemplate_template
        self.out_clustertemplate_name = out_clustertemplate_name
        self.out_ansible_hosts_name = out_ansible_hosts_name
        self.out_ansible_vars_name = out_ansible_vars_name
        self.ansible_sudo_pass = ansible_sudo_pass


    def get_ini(self, file):
        ini = configparser.ConfigParser()
        ini.read(file)
        return ini

    def return_hosts(self):
        try:
            with open(self.start_hosts_file) as f:
                return yaml.safe_load(f)
        except FileNotFoundError:
            logging.critical(f'Ошибка, не нашел стартовый файл {self.start_hosts_file}')
            sys.exit(1)


    def get_host_groups(self):
        host_groups = {}
        for host in self.return_hosts():
            group = host['host_group']
            if 'host_group' in host:
                if not group in host_groups:
                    host_groups[group] = {}
                host_groups[group][host['name']] = host['name']
        result_host_groups = {}
        for group, hosts in host_groups.items():
            result_host_groups[group] = list(hosts)
        return result_host_groups


    def get_ansible_groups(self):
        groups = {}
        for host in self.return_hosts():
            # print(host)
            # Магия, чтоб одну группу всё же превратить в список,
            ansible_groups = host['ansible_groups']
            # из ямла прилетатет строка, если группа одна.
            if isinstance(ansible_groups, str):
                ansible_groups = []
                ansible_groups.append(host['ansible_groups'])
            if ansible_groups:   # Если нет груп ансибл, то и хост не нужен в инвентори
                host_name = host['name']
                for group in ansible_groups:
                    if not group in groups:
                        groups[group] = {}
                    groups[group][host_name] = host


        return groups

    def generate(self, template_file=''):
        file_loader = FileSystemLoader(self.templates_folder)
        template = Environment(loader=file_loader).get_template(template_file)
        result = template.render(content=self.return_content())
        return result


    def save_to_file(self, file='', data=''):
        out_file = f'{self.path_out_file}/{file}'
        try:
            with open(out_file, "w") as f:
                f.write(data)
        except FileNotFoundError:
            logging.critical(
                f'Ошибка, нет папки {self.path_out_file}, не смог сохранить {file}')


    def return_content(self):
        configs = {}
        config = self.get_ini(self.config)
        for section in config:
            for param, param_val in config[section].items():
                if not section in configs:
                    configs[section] = {}
                configs[section][param] = param_val

        content = {
            'metadata':  {'cluster_name': self.cluster_name},
            'auth': {
                'super_secret_pass': self.super_secret_pass,
                'service_user_pass': self.service_user_pass,
                'ansible_ssh_pass': self.ansible_ssh_pass,
                'ansible_sudo_pass': self.ansible_sudo_pass
                },
            'configs': configs,
            'ansible_groups': self.get_ansible_groups(),
            'host_groups': self.get_host_groups()
        }
        return content
    
    def generate_hosts(self):
        ansible_hosts = self.generate(template_file=self.ansible_hosts_template)
        self.save_to_file(data=ansible_hosts, file=self.out_ansible_hosts_name)

    def generate_clustertemplate(self):
        clustertemplate = self.generate(
            template_file=self.clustertemplate_template)
        self.save_to_file(data=clustertemplate,
                          file=self.out_clustertemplate_name)
    
    def generate_vars(self):
        ansible_vars = self.generate(template_file=self.ansible_vars_template)
        self.save_to_file(data=ansible_vars, file=self.out_ansible_vars_name)


    def generate_blueprint(self):
        env_prefix = self.return_content()['configs']['global']['env_prefix']
        domain = self.return_content()['configs']['global']['domain']
        cl_type = None
        if 'cluster_type' in self.return_content()['configs']['global']:
            cl_type = self.return_content()['configs']['global']['cluster_type']
        if not cl_type:
            logging.critical(f'Не задан параметр "cluster_type" в секции "global"')
            exit(1)
        blueprint_template = f'{self.templates_folder}/blueprint_{cl_type}.json'
        cmds = [
            f'cp {blueprint_template} {self.path_out_file}/blueprint.json',
            f'sed -i "s/___domain_name___/.{domain}/g" {self.path_out_file}/blueprint.json',
            f'sed -i "s/___password___/{self.super_secret_pass}/g" {self.path_out_file}/blueprint.json'
            ]
        for cmd in cmds:
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            process.wait()

        # Ключ должен совпадать с именем группы host_group в стартовом файле
        # 
        template_groups = [
            {'cluster_manager': '___cluster_manager___'},
            {'etl_control_03': '___etl_control_03___'},
            {'namenode_1': '___namenode_1___'},
            {'service_02': '___service-02___'},
            {'security_01': '___security-01___'},
            {'namenode': '___namenode___'},
            {'datanode': False},
            {'control': False},
            {'kafka': False}
            ]
        c = 0
        c_tmp_group_key = ''
        for group, hosts in self.get_host_groups().items():
            if group:
                for temp_group in template_groups:
                    for tmp_group_key, tmp_group_val in temp_group.items():
                        if tmp_group_key in group:
                            for host in hosts:
                                if env_prefix:
                                    host = f'{env_prefix}-{host}'
                                cmd = f'sed -i "s/{tmp_group_val}/{host}/g" {self.path_out_file}/blueprint.json'
                                if not tmp_group_val:
                                    if tmp_group_key != c_tmp_group_key:
                                        c_tmp_group_key = tmp_group_key
                                        c = 0
                                    c += 1
                                    cmd = f'sed -i "s/___{tmp_group_key}-0{c}___/{host}/g" {self.path_out_file}/blueprint.json'
                                process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
                                process.wait()
