from build_files import BuildFiles
import sys
import os
import re
import hashlib
import logging
import pathlib
import socket
import requests

LEVEL = 'INFO'
FORMAT = "[{levelname:^10s}] - {message:s}"
logging.basicConfig(format=FORMAT, style='{', level=LEVEL)

current_path = str(pathlib.Path(__file__).parent.absolute())


start_hosts_file = 'start_hosts.yml'
config_file = 'config.ini'
path_out_file = 'files_for_installer'

def log_errors(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logging.error(f"{e}")
            exit(1)
    return wrapper

@log_errors
def check_passord(password):
    password_name = 'PLATFORM_INSTALLER_ADMIN_PASSWORD'
    if password == '':
        raise Exception(f'Ошибка, пароль ({password_name}) не может быть пустым!')
    required_alphabet_small = re.compile(r'[a-z]')
    required_alphabet_big = re.compile(r'[A-Z]')
    required_numeric = re.compile(r'[0-9]')
    required_special = re.compile(r'[~!^*_+,().;\-]')
    valid_pass_regex = re.compile(r'^[0-9A-Za-z~!^*_+,().;\-]+$')

    exception_massages = []
    if not required_alphabet_small.search(password):
        exception_massages.append(f'Ошибка, пароль ({password_name}) должен содержать как минимум одну строчную букву')

    if not required_alphabet_big.search(password):
        exception_massages.append(f'Ошибка, пароль ({password_name}) должен содержать как минимум одну заглавную букву')

    if not required_numeric.search(password):
        exception_massages.append(f'Ошибка, пароль ({password_name}) должен содержать как минимум одну цифру')

    if not required_special.search(password):
        exception_massages.append(f'Ошибка, пароль ({password_name}) должен содержать как минимум один спецсимвол из разрешённого списка')

    if len(password) < 8:
        exception_massages.append(f'Ошибка, пароль ({password_name}) должен быть не менее 8 символов')

    if not valid_pass_regex.search(password):
        exception_massages.append(f'Ошибка, в пароле ({password_name}) есть неподдерживаемый спецсимвол')
    
    if exception_massages:
        for excpt in exception_massages:
            logging.error(excpt)
        raise Exception('Выход')

@log_errors
def check_config_file(build):
    exception_massages = []
    check_params = {
        'ansible': ['become_method', 'become_user', 'ansible_user'],
        'global': ['domain', 'env_prefix', 'cluster_type'],
        'nexus': ['raw', 'yum', 'docker']
    }
    configs = build.return_content()['configs']
    for section, params in check_params.items():
        if not section in configs:
            raise Exception(f"Не определена секция '{section}' в '{config_file}'!")
        for param in params:
            if not param in configs[section]:
                exception_massages.append(f"Не задан параметр '{param}' в секции '{section}', файла '{config_file}'")
    if exception_massages:
        for excpt in exception_massages:
            logging.error(excpt)
        raise Exception('Выход')
    return True

@log_errors
def download_configs():
    logging.info('Попытка скачать конфигурационные файлы')
    hostname = socket.gethostname().split('-installer')[0]
    # hostname = 'rnd-infra-1'.split('-installer')[0]
    logging.debug(f'{hostname=}')
    # https://nexus.rzd/repository/cs-khd-gvc/contours/rzd-bdp/start_hosts.yml
    # https://nexus.corp.myservices.digital/repository/khd-raw/contours/rnd-infra-1/start_hosts.yml
    nexus = {
        'rzd': 'https://nexus.rzd/repository/cs-khd-gvc',
        'rnd': 'https://nexus.corp.myservices.digital/repository/khd-raw'
        }

    nexus_prefix = hostname.split('-')[:1][0]

    if not nexus_prefix in nexus:
        raise Exception(f'Ошибка неподдерживаемое имя контура {hostname}')

    download_files = [start_hosts_file, config_file]
    for download in download_files:
        url = f'{nexus[nexus_prefix]}/contours/{hostname}/{download}'
        logging.debug(f'{url=}')
        r = requests.get(url, allow_redirects=True, verify=False)
        if r.status_code != 200:
            raise Exception(f'Ошибка скачивания файла {download}')
        open(f'{current_path}/{download}', 'wb').write(r.content)
        logging.info(f"Успешно скачан файл '{download}'")
    return True


for file in [start_hosts_file, config_file]:
    if not os.path.isfile(f'{current_path}/{file}'):
        logging.critical(f'Ошибка, не найден файл {file}')
        download_configs()

@log_errors
def main():
    ssh_password = os.environ.get('PLATFORM_INSTALLER_SSH_PASSWORD')
    if not ssh_password:
        raise Exception('Ошибка, не найдена переменная окружения PLATFORM_INSTALLER_SSH_PASSWORD')

    password = os.environ.get('PLATFORM_INSTALLER_ADMIN_PASSWORD')
    if not password:
        raise Exception('Ошибка, не найдена переменная окружения PLATFORM_INSTALLER_ADMIN_PASSWORD')

    # Проверка пароля
    check_passord(password)

    # BDP-4063 Пароль для service_user это md5 от переменной password
    service_user_pass = hashlib.md5(f'service_user{password}'.encode('utf-8')).hexdigest()

    # А вдруг папки нет для результатов работы?
    os.makedirs(f'{current_path}/files_for_installer', exist_ok = True)
    build = BuildFiles(
        start_hosts_file = start_hosts_file,
        config_file = config_file,
        path_out_file = path_out_file,
        super_secret_pass = password,
        service_user_pass = service_user_pass,
        ansible_ssh_pass = ssh_password,
        ansible_sudo_pass = ssh_password
    )

    # Проверка конфиг файла
    check_config_file(build)

    # Генератор списка, который ищет хосты и плохие символы
    re_search = "([^-^a-z\d^A-Z])"  # Ищет всё, что не буквы не цифры и не -
    bad_hosts = [
        {host['name']: re.findall(re_search, host['name'])} 
        for host in build.return_hosts()
        if re.findall(re_search, host['name'])
        ]

    # Если что-то попало в генератор списка с невалидными именами хостов
    if bad_hosts:
        logging.info('\n##########################')
        for host_and_symbol in bad_hosts:
            for host, symbol in host_and_symbol.items():
                msg = f'Недопустимый символ "{symbol[0]}" в имени хоста: {host}'
                if len(symbol) > 1:
                    l_symbol = ' '.join(symbol)
                    msg = f'Недопустимые символы "{l_symbol}" в имени хоста: {host}'
                raise Exception(msg)

    # Для отладки, данный скрипт можно запустить без последствий (генерация файлов), с параметром 'pass'
    if 'pass' not in sys.argv:
        logging.info('Создаю файлы конфигурации')
        build.generate_hosts()
        build.generate_vars()
        build.generate_clustertemplate()
        build.generate_blueprint()
        logging.info(f"Установка будет выполнена от пользователя '{build.return_content()['configs']['ansible']['ansible_user']}'")
        logging.info(f"Все файлы созданы, указанный пароль сохранен.")


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("\nВыход")
        exit(0)
